import omni.ext
import omni.ui as ui
import os
import sys
import weakref
import threading
import asyncio
import traceback
import gc
from enum import Enum

from . import scenario
from .scenario import (
    ModulusOVContainer,
    ModulusOVRow,
    ModulusOVSpacer,
    ModulusOVElement,
    ModulusOVText,
    ModulusOVButton,
    ModulusOVToggle,
    ModulusOVFloatSlider,
    ModulusOVIntSlider,
    ModulusOVSlider,
    ModulusOVField,
    ModulusOVStrField,
    ModulusOVFloatField,
    ModulusOVIntField,
    ModulusOVComboBox,
    ModulusOVProgressBar,
)

sys.path.append(os.path.dirname(__file__))

WINDOW_MENU = "Window/Modulus/Modulus Scenario Parameters"


class SliderType(Enum):
    FLOAT = 1
    INT = 2


class FieldType(Enum):
    STRING = 1
    FLOAT = 2
    INT = 3


# Any class derived from `omni.ext.IExt` in top level module
# (defined in `python.modules` of `extension.toml`) will be
# instantiated when extension gets enabled and `on_startup(ext_id)`
# will be called. Later when extension gets disabled
# on_shutdown() is called.
class ModulusExt(omni.ext.IExt):
    scenario_registry = []
    window = None

    def add_scenario_ui(self, s: scenario.ModulusOVScenario):
        for c in s.containers.values():
            if c.empty:
                continue
            else:
                self.add_container(c)
                ui.Spacer(height=10)

        s.busyLabel = ui.Label(
            "Operation in progress...", alignment=ui.Alignment.H_CENTER
        )
        s.busyLabel.visible = False

    def add_container(self, c: ModulusOVContainer):
        with ui.CollapsableFrame(
            title=c.name, height=c.height, collapsed=c.collapsed, style=c.style
        ):
            # Loop through elements in Modulus OV container
            with ui.VStack(height=0, spacing=1, style={"VStack": {"margin_width": 10}}):
                if c.desc:
                    ui.Label(c.desc, word_wrap=True)
                    ui.Spacer(height=1)

                for e in c.elements:
                    self.add_element(e)
                    ui.Spacer(height=1)

    def add_element(self, e: ModulusOVElement):
        print(e)
        if isinstance(e, ModulusOVRow) and not e.empty:
            if e.desc:
                ui.Label(e.desc, word_wrap=True)
            with ui.HStack(
                height=e.height,
                width=ui.Percent(100),
                column_count=len(e.elements),
                spacing=e.spacing,
                alignment=ui.Alignment.CENTER,
            ):
                for e_row in e.elements:
                    e_row.width = ui.Percent(100 / len(e.elements))
                    self.add_element(e_row)
        elif isinstance(e, ModulusOVSpacer):
            self.add_spacer(e)
        elif isinstance(e, ModulusOVText):
            self.add_text(e)
        elif isinstance(e, ModulusOVButton):
            self.add_button(e)
        elif isinstance(e, ModulusOVToggle):
            self.add_toggle(e)
        elif isinstance(e, ModulusOVFloatSlider):
            self.add_slider(e, SliderType.FLOAT)
        elif isinstance(e, ModulusOVIntSlider):
            self.add_slider(e, SliderType.INT)
        elif isinstance(e, ModulusOVStrField):
            self.add_field(e, FieldType.STRING)
        elif isinstance(e, ModulusOVFloatField):
            self.add_field(e, FieldType.FLOAT)
        elif isinstance(e, ModulusOVIntField):
            self.add_field(e, FieldType.INT)
        elif isinstance(e, ModulusOVComboBox):
            self.add_combo(e)
        elif isinstance(e, ModulusOVProgressBar):
            self.add_progress(e)
        else:
            raise ValueError(
                f"Invalid UI element: {e}, {type(e).__name__}, {e.__class__}, {ModulusOVStrField}"
            )

    def add_spacer(self, e: ModulusOVSpacer):
        ui.Spacer(width=e.width, height=e.height)

    def add_text(self, e: ModulusOVText):
        label = ui.Label(
            e.desc,
            alignment=e.alignment,
            word_wrap=True,
            style=e.style,
        )
        e.ui = label

    def add_button(self, e: ModulusOVButton):
        button = ui.Button(
            e.name,
            tooltip=e.desc,
            clicked_fn=lambda s=self.scenario_registry[0](), e=e: self.invoke(
                s, e
            ),  # TODO: Fix this jank
            visible=e.visible,
            height=e.height,
            width=e.width,
        )
        e.ui = button

    def add_toggle(self, e: ModulusOVButton):
        with ui.HStack():
            ui.Label(e.name, tooltip=e.desc, width=ui.Percent(50))
            ui.Line(
                width=ui.Percent(45), style={"color": 0xFF666666, "margin_width": 10}
            )
            box = omni.ui.CheckBox(width=ui.Percent(5))
            box.model.set_value(e.value)
            box.model.add_value_changed_fn(
                lambda m=box.model, p=e: self.update_toggle(m, p)
            )

    def add_slider(self, e: ModulusOVSlider, stype: SliderType = SliderType.FLOAT):
        def _bounds_changed_float(s, new_bounds):
            minval, maxval = new_bounds
            minval = float(minval)
            maxval = float(maxval)
            s.min = minval
            s.max = maxval
            s.model.set_min(minval)
            s.model.set_max(maxval)

        def _bounds_changed_int(s, new_bounds):
            minval, maxval = new_bounds
            minval = int(minval)
            maxval = int(maxval)
            s.min = minval
            s.max = maxval
            s.model.set_min(minval)
            s.model.set_max(maxval)

        with ui.HStack():
            ui.Label(e.name, tooltip=e.desc, width=ui.Percent(e.label_width))
            ui.Spacer(width=ui.Percent(5))

            if stype == SliderType.FLOAT:
                slider = ui.FloatSlider(width=e.width)
            elif stype == SliderType.INT:
                slider = ui.IntSlider(width=e.width)
            else:
                raise ValueError("Invalid slider type")

            slider.model.set_value(e.value)
            if stype == SliderType.FLOAT:
                slider.model.add_value_changed_fn(
                    lambda m=slider.model, p=e: self.update_parameter(
                        m, p, SliderType.FLOAT
                    )
                )
                _bounds_changed_float(slider, e.bounds)
                e.add_bounds_changed_fn(lambda b, s=slider: _bounds_changed_float(s, b))
            elif stype == SliderType.INT:
                slider.model.add_value_changed_fn(
                    lambda m=slider.model, p=e: self.update_parameter(
                        m, p, SliderType.INT
                    )
                )
                _bounds_changed_int(slider, e.bounds)
                e.add_bounds_changed_fn(lambda b, s=slider: _bounds_changed_int(s, b))
            ui.Spacer(width=5)

    def add_field(self, e: ModulusOVField, ftype: FieldType = FieldType.STRING):
        with ui.HStack():
            ui.Label(e.name, tooltip=e.desc, width=ui.Percent(e.label_width))
            ui.Spacer(width=ui.Percent(5))
            if ftype == FieldType.STRING:
                field = ui.StringField(height=e.height, width=e.width)
                field.model.add_value_changed_fn(
                    lambda m=field.model, p=e: self.update_field(m, p, FieldType.STRING)
                )
            elif ftype == FieldType.INT:
                field = ui.IntField(height=e.height, width=e.width)
                field.model.add_value_changed_fn(
                    lambda m=field.model, p=e: self.update_field(m, p, FieldType.INT)
                )
            elif ftype == FieldType.FLOAT:
                field = ui.FloatField(height=e.height, width=e.width)
                field.model.add_value_changed_fn(
                    lambda m=field.model, p=e: self.update_field(m, p, FieldType.FLOAT)
                )
            else:
                raise ValueError("Invalid slider type")
            field.model.set_value(e.value)

    def add_combo(self, e: ModulusOVComboBox):
        with ui.HStack():
            ui.Label(e.name, tooltip=e.desc, width=ui.Percent(e.label_width))
            ui.Spacer(width=ui.Percent(5))
            labels = [str(o) for o in e.options]
            box = ui.ComboBox(e.default_index, *labels, width=e.width, height=e.height)
            box.model.add_item_changed_fn(lambda m, i, p=e: self.update_combo(m, p))

    def add_progress(self, e: ModulusOVProgressBar):
        bar = ui.ProgressBar(
            width=e.width,
            height=e.height,
            tooltip=e.desc,
            style=e.style,
        )
        bar.model.set_value(e.value)
        e.ui = bar.model

    def invoker_thread(
        self, s: scenario.ModulusOVScenario, a: scenario.ModulusOVButton
    ):
        if a.update_func is None:
            return
        try:
            with s.mutex:
                a.update_func()
        except:
            print("Error during action invocation.")
            raise
        finally:
            print("Invoked done")

    def invoke(self, s: scenario.ModulusOVScenario, a: scenario.ModulusOVButton):
        if not hasattr(s, "mutex"):
            s.mutex = threading.Lock()
        if a.run_in_main_thread:
            self.invoker_thread(s, a)
            return

        res = s.mutex.acquire(timeout=0.01)
        if res:
            try:
                threading.Thread(
                    target=self.invoker_thread, args=(s, a), kwargs={}
                ).start()
            except:
                print("Error during scenario action.")
                raise
            s.mutex.release()
        else:
            print("Unable to acquire scenario mutex.")

    def update_parameter(self, m, p, stype: SliderType = SliderType.FLOAT):
        if not hasattr(self, "mutex"):
            self.mutex = threading.Lock()
        try:
            with self.mutex:
                if stype == SliderType.FLOAT:
                    setattr(p, "value", m.get_value_as_float())
                elif stype == SliderType.INT:
                    setattr(p, "value", m.get_value_as_int())
        except:
            print("Error during scenario action.")
            raise

    def update_toggle(self, m, p):
        if not hasattr(self, "mutex"):
            self.mutex = threading.Lock()
        try:
            with self.mutex:
                setattr(p, "value", m.get_value_as_bool())
        except:
            print("Error during scenario action.")
            raise

    def update_combo(self, m, p):
        if not hasattr(self, "mutex"):
            self.mutex = threading.Lock()
        try:
            with self.mutex:
                setattr(p, "value", m.get_item_value_model().as_int)
        except:
            print("Error during scenario action.")
            raise

    def update_field(self, m, p, ftype: FieldType = FieldType.STRING):
        if not hasattr(self, "mutex"):
            self.mutex = threading.Lock()
        try:
            with self.mutex:
                if ftype == FieldType.STRING:
                    setattr(p, "value", m.get_value_as_string())
                elif ftype == FieldType.FLOAT:
                    setattr(p, "value", m.get_value_as_float())
                elif ftype == FieldType.INT:
                    setattr(p, "value", m.get_value_as_int())
        except:
            print("Error during scenario action.")
            raise

    def update_ui(self):
        print("[modulus_ext.ui] Update UI called.")
        print("registry: {}".format(self.scenario_registry))
        # Added to clear GUI when developing
        ModulusExt.window.frame.clear()
        with ModulusExt.window.frame:
            with ui.ScrollingFrame():
                with ui.VStack():
                    for s_weakref in ModulusExt.scenario_registry:
                        try:
                            ui.Spacer(height=5)
                            self.add_scenario_ui(s_weakref())
                            ui.Spacer(height=5)
                        except Exception as e:
                            ui.Label("Modulus OV UI Error", tooltip="error")
                            traceback.print_exc()

    # ext_id is current extension id. It can be used with extension manager
    # to query additional information, like where
    # this extension is located on filesystem.
    def on_startup(self, ext_id):
        print("[modulus_ext.ui] Starting up.")
        ModulusExt.window = ui.Window(
            "Modulus",
            width=300,
            height=300,
            dockPreference=ui.DockPreference.RIGHT_BOTTOM,
            visible=False,
        )

        editor_menu = omni.kit.ui.get_editor_menu()

        if editor_menu:
            self._menu = editor_menu.add_item(
                WINDOW_MENU,
                self._on_menu_click,
                toggle=True,
                value=self._is_window_visible(),
            )
        ModulusExt.window.set_visibility_changed_fn(self._visibility_changed_fn)
        ModulusExt.window.deferred_dock_in("Property")
        self.update_ui()
        self._dock_task = None
        self._dock_eventually()
        # subscribe_modulus_ext.ui_scenarios(self.on_scenarios_updated)

    def _dock_eventually(self):
        if self._dock_task is not None:
            self._dock_task.cancel()
        self._dock_task = asyncio.ensure_future(self._dock())

    async def _dock(self):
        prop = ui.Workspace.get_window("Property")
        if prop:
            # Need to be a few frames in before dock can occur.
            await omni.kit.app.get_app().next_update_async()
            await omni.kit.app.get_app().next_update_async()
            await omni.kit.app.get_app().next_update_async()
            await omni.kit.app.get_app().next_update_async()
        ModulusExt.window.dock_in(prop, ui.DockPosition.RIGHT)
        self._dock_task = None
        ModulusExt.window.visible = True

    def _is_window_visible(self):
        return ModulusExt.window.visible

    def _visibility_changed_fn(self, visible):
        omni.kit.ui.get_editor_menu().set_value(WINDOW_MENU, visible)

    def _on_menu_click(self, *args):
        ModulusExt.window.visible = not ModulusExt.window.visible

    def register_scenario(self, s: scenario.ModulusOVScenario):
        ModulusExt.scenario_registry.append(weakref.ref(s))
        print('[modulus_ext.ui] Registered scenario "{}"'.format(s.name))
        self.update_ui()

    def unregister_scenario(self, s: scenario.ModulusOVScenario):
        while weakref.ref(s) in ModulusExt.scenario_registry:
            ModulusExt.scenario_registry.remove(weakref.ref(s))
            print(weakref.ref(s))
        print('[modulus_ext.ui] Unregistered scenario "{}"'.format(s.name))
        gc.collect()
        self.update_ui()

    def on_shutdown(self):
        print("[modulus_ext.ui] Shutting down.")
        ModulusExt.window = None
