import omni.ui as ui
from typing import Callable
from typing import List
from typing import Tuple
from typing import TypeVar
from typing import Union
from dataclasses import dataclass, field

T = TypeVar("T")
S = TypeVar("S", int, float)
F = TypeVar("F", int, float, str)

WIDGET_HEIGHT = 0


class ModulusOVElement:
    def __init__(
        self,
        name: str,
        desc: str = "Modulus OV Element",
        visible: bool = True,
        run_in_main_thread: bool = True,  # TODO: Make out of thread more realiable
    ):
        self.name = name
        self.desc = desc
        self.visible = visible
        self.run_in_main_thread = run_in_main_thread
        self.ui = None


class ModulusOVContainer(ModulusOVElement):
    def __init__(
        self,
        name: str,
        desc: str = "Modulus OV Container",
        collapsed: bool = False,
        height: int = WIDGET_HEIGHT,
        elements: List[ModulusOVElement] = [],
    ):
        super().__init__(name=name, desc=desc)
        self.collapsed = collapsed
        self.height = height
        self.elements = []
        self.style = {
            "CollapsableFrame": {
                "background_color": 0xFF343432,
                "secondary_color": 0xFF343432,
                "color": 0xFFAAAAAA,
                "border_radius": 4.0,
                "border_color": 0x0,
                "border_width": 0,
                "font_size": 14,
                "padding": 5,
            },
            "HStack::header": {"margin": 10},
            # "CollapsableFrame:hovered": {"secondary_color": 0xFF3A3A3A},
            # "CollapsableFrame:pressed": {"secondary_color": 0xFF343432},
        }

    def add(self, element: ModulusOVElement):
        self.elements.append(element)

    def __iadd__(self, element: ModulusOVElement):
        self.elements.append(element)

    @property
    def empty(self):
        return len(self.elements) == 0


class ModulusOVRow(ModulusOVElement):
    def __init__(
        self,
        elements: List[ModulusOVElement] = [],
        desc: Union[str, None] = None,
        spacing: int = 5,
    ):
        super().__init__(name="row", desc=desc)
        self.height = 30
        self.elements = elements
        self.spacing = spacing

    def add(self, element: ModulusOVElement):
        self.elements.append(element)

    def __iadd__(self, element: ModulusOVElement):
        self.elements.append(element)

    @property
    def empty(self):
        return len(self.elements) == 0


class ModulusOVSpacer(ModulusOVElement):
    def __init__(
        self,
        width: int = 0,
        height: int = 0,
    ):
        super().__init__(name="spacer")
        self.height = height
        self.width = width


class ModulusOVText(ModulusOVElement):
    def __init__(
        self,
        desc: str = "Modulus OV Test",
        alignment=ui.Alignment.LEFT,
    ):
        super().__init__(name="Text block", desc=desc)
        self.alignment = alignment
        self.style = {"color": 0xFF00B976, "margin_height": 5, "font_size": 16}

    @property
    def text(self):
        return self.desc

    @text.setter
    def text(self, value: str):
        self.desc = value
        # Update element if assigned
        if not self.ui is None:
            self.ui.text = value


class ModulusOVButton(ModulusOVElement):
    def __init__(
        self,
        name: str,
        update_func: Callable[[], None],
        desc: str = "Modulus OV Button",
        height: int = WIDGET_HEIGHT,
        width: int = ui.Percent(100),
    ):
        super().__init__(name=name, desc=desc)
        self.update_func = update_func
        self.height = height
        self.width = width

    @property
    def text(self):
        return self.name

    @text.setter
    def text(self, value: str):
        self.name = value
        # Update element if assigned
        if not self.ui is None:
            self.ui.text = value


class ModulusOVToggle(ModulusOVElement):
    def __init__(
        self,
        name: str,
        update_func: Callable[[], None],
        desc: str = "Modulus OV Toggle",
        default_value: bool = False,
        width: int = ui.Percent(50),
    ):
        super().__init__(name=name, desc=desc)
        self.update_func = update_func
        self.width = width
        self._value = default_value

    # Only supports 1 way update (cannot control value in script, only on init)
    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value
        if self.update_func:
            self.update_func(value)


class ModulusOVSlider(ModulusOVElement):
    def __init__(
        self,
        name: str,
        default_value: S,
        bounds: Tuple[S, S],
        update_func: Callable[[S], None],
        desc: str = "Modulus OV Toggle",
    ):
        super().__init__(name=name, desc=desc)
        self.update_func = update_func
        self.bounds = bounds
        self.width = ui.Percent(64)
        self.label_width = ui.Percent(30)
        self._value = default_value
        self._bounds_changed_fns = []

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value
        if self.update_func:
            self.update_func(value)

    def update_bounds(self, bounds: Tuple[S, S]):
        self._bounds = bounds
        for fn in self._bounds_changed_fns:
            fn(self._bounds)

    def add_bounds_changed_fn(self, fn: Callable[[Tuple[S, S]], None]):
        self._bounds_changed_fns.append(fn)


class ModulusOVFloatSlider(ModulusOVSlider):
    def __init__(
        self,
        name: str,
        default_value: float,
        bounds: Tuple[float, float],
        update_func: Callable[[float], None],
        desc: str = "Modulus OV Float Slider",
    ):
        super().__init__(
            name=name,
            default_value=default_value,
            bounds=bounds,
            update_func=update_func,
            desc=desc,
        )


class ModulusOVIntSlider(ModulusOVSlider):
    def __init__(
        self,
        name: str,
        default_value: int,
        bounds: Tuple[int, int],
        update_func: Callable[[int], None],
        desc: str = "Modulus OV Int Slider",
    ):
        super().__init__(
            name=name,
            default_value=default_value,
            bounds=bounds,
            update_func=update_func,
            desc=desc,
        )

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = int(value)
        if self.update_func:
            self.update_func(int(value))


class ModulusOVField(ModulusOVElement):
    def __init__(
        self,
        name: str,
        default_value: F,
        desc: str = "",
        update_func: Callable[[int], None] = None,
        width: int = ui.Percent(49),
        height: int = 20,
    ):
        super().__init__(name=name, desc=desc)
        self.update_func = update_func
        self.height = height
        self.width = width
        self.label_width = ui.Percent(45)
        self._value = default_value

    @property
    def value(self):
        return self._value


class ModulusOVStrField(ModulusOVField):
    def __init__(
        self,
        name: str,
        desc: str = "Modulus OV String Field",
        default_value: str = "",
        update_func: Callable[[str], None] = None,
        height: int = 20,
    ):
        super().__init__(
            name=name,
            default_value=default_value,
            desc=desc,
            height=height,
            update_func=update_func,
        )

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value
        if self.update_func:
            self.update_func(str(value))


class ModulusOVFloatField(ModulusOVField):
    def __init__(
        self,
        name: str,
        desc: str = "Modulus OV Float Field",
        default_value: float = 0,
        update_func: Callable[[float], None] = None,
        height: int = 20,
    ):
        super().__init__(
            name=name,
            default_value=default_value,
            desc=desc,
            height=height,
            update_func=update_func,
        )

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value
        if self.update_func:
            self.update_func(float(value))


class ModulusOVIntField(ModulusOVField):
    def __init__(
        self,
        name: str,
        desc: str = "Modulus OV Int Field",
        default_value: int = 0,
        update_func: Callable[[int], None] = None,
        height: int = 20,
    ):
        super().__init__(
            name=name,
            default_value=default_value,
            desc=desc,
            height=height,
            update_func=update_func,
        )

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value
        if self.update_func:
            self.update_func(int(value))


class ModulusOVComboBox(ModulusOVElement):
    def __init__(
        self,
        name: str,
        options: List,
        desc: str = "Modulus OV Combo Box",
        update_func: Callable[[int], None] = None,
        default_index: int = 0,
    ):
        super().__init__(
            name=name,
            desc=desc,
        )
        self.options = options
        self.width = ui.Percent(49)
        self.label_width = ui.Percent(45)
        self.height = 20
        self.update_func = update_func
        self.default_index = default_index
        self._value = self.options[default_index]

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, index: int):
        self._value = self.options[index]
        if self.update_func:
            self.update_func(self.options[index])


class ModulusOVProgressBar(ModulusOVElement):
    def __init__(
        self,
        default_value: float = 1.0,
        desc: str = "Modulus OV Progress Bar",
        height: int = WIDGET_HEIGHT,
        width: int = ui.Percent(100),
    ):
        super().__init__(name="Progress Bar", desc=desc)
        self.height = height
        self.width = width
        self._value = default_value
        self.inference_scale = 1.0
        self.style = {
            "color": 0xFFE79E55,
            "margin_height": 5,
            "border_radius": 2.0,
        }
        self._prev_step = 0

    # Special call to be used in Modulus inferencers
    def inference_step(self, value: float):
        self.value = self.inference_scale * value + self._value - self._prev_step
        self._prev_step = self.inference_scale * value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value: float):
        self._value = max(0, min(value, 1.0))
        # Update element if assigned
        if not self.ui is None:
            self.ui.set_value(self.value)


class ModulusOVScenario(object):
    """Modulus scenario base class. All scenarios should inherit this

    Parameters
    ----------
    name : str
        Name of scenario
    desc : Union[str, None], optional
        Optional discription fo the scenario, by default None
    """

    def __init__(
        self,
        name: str,
        desc: Union[str, None] = None,
    ):
        from .. import ModulusExt

        mExt = ModulusExt()

        self.name = name
        self.desc = desc

        self.default_container = ModulusOVContainer(
            name=self.name,
            desc=self.desc,
            elements=[],  # Somehow this can persist between scenario reloads if not set
        )
        self.containers = {"default": self.default_container}
        self.processing = False

    def add(self, element: ModulusOVElement, key: Union[str, None] = None):
        if key is None:
            key = next(iter(self.containers))
        assert key in self.containers, f"invalid contrainer name: {key}"

        self.containers[key].add(element)

    def add_container(
        self, name: str, desc: Union[str, None] = None, key: Union[str, None] = None
    ):
        if key is None:
            key = f"container{len(self.containers)}"
        self.containers[key] = ModulusOVContainer(name=name, desc=desc, elements=[])

    def register(self):
        from .. import ModulusExt

        mExt = ModulusExt()
        mExt.register_scenario(self)

    def action(func):
        def wrapper(self, *args, **kwargs):
            if not self.processing:
                self.processing = True
                func(self, *args, **kwargs)
                self.processing = False
            else:
                print("Modulus extension already processing an action")

        return wrapper

    def __del__(self):
        del self.containers
        from .. import ModulusExt

        mExt = ModulusExt()
        mExt.unregister_scenario(self)
